from __future__ import annotations

import itertools
from typing import cast

import cirq
import numpy as np

from supermarq import stabilizers
from supermarq.benchmark import Benchmark


class MerminBell(Benchmark):
    """The Mermin-Bell benchmark is a test of a quantum computer's ability to exploit purely quantum
    phenomemna such as superposition and entanglement. It is based on the famous Bell-inequality
    tests of locality. Performance is based on a QPU's ability to prepare a GHZ state and measure
    the Mermin operator.
    """

    def __init__(self, num_qubits: int) -> None:
        """Initializes a `MerminBell`.

        Args:
            num_qubits: The number of qubits.
        """
        self.num_qubits = num_qubits
        self.qubits = cirq.LineQubit.range(self.num_qubits)

        self.mermin_operator = self._mermin_operator(self.num_qubits)
        self.stabilizer, self.pauli_basis = stabilizers.construct_stabilizer(
            self.num_qubits, self.mermin_operator
        )

    def circuit(self) -> cirq.Circuit:
        """The Mermin-Bell circuit, simultaneously measuring Mermin terms in a GHZ circuit.

        Returns:
            The Mermin-Bell `cirq.Circuit`.
        """
        circuit = cirq.Circuit()

        # Create a GHZ state
        circuit.append(cirq.rx(-np.pi / 2).on(self.qubits[0]))
        for i in range(self.num_qubits - 1):
            circuit.append(cirq.CNOT(self.qubits[i], self.qubits[i + 1]))

        # Simultaneously measure all terms in the Mermin operator
        measurement_circuit = self._get_measurement_circuit()
        circuit.append(measurement_circuit.get_circuit())

        return circuit

    def score(self, counts: dict[str, float]) -> float:
        """Compute the score for the N-qubit Mermin-Bell benchmark.

        This function assumes the regular big endian ordering of bitstring results.

        Args:
            counts: A dictionary containing the measurement counts from circuit execution.

        Returns:
            The score for the Mermin-Bell benchmark score.
        """
        # Store the conjugation rules for H, S, CX, CZ, SWAP in dictionaries. The keys are
        # the pauli strings to be conjugated and the values are the resulting pauli strings
        # after conjugation.
        # The typing here was added to satisfy mypy. Declaring this dict without the explicit
        # typing gets created as Dict[EigenGate, Dict[str, str]], but iterating through a
        # cirq.Circuit and passing op.gate as the key yields type Optional[Gate].
        conjugation_rules: dict[cirq.Gate | None, dict[str, str]] = {
            cirq.ops.H: {"I": "I", "X": "Z", "Y": "-Y", "Z": "X"},
            cirq.ops.S: {"I": "I", "X": "Y", "Y": "-X", "Z": "Z"},
            cirq.ops.CNOT: {
                "II": "II",
                "IX": "IX",
                "XI": "XX",
                "XX": "XI",
                "IY": "ZY",
                "YI": "YX",
                "YY": "-XZ",
                "IZ": "ZZ",
                "ZI": "ZI",
                "ZZ": "IZ",
                "XY": "YZ",
                "YX": "YI",
                "XZ": "-YY",
                "ZX": "ZX",
                "YZ": "XY",
                "ZY": "IY",
            },
            cirq.ops.CZ: {
                "II": "II",
                "IX": "ZX",
                "XI": "XZ",
                "XX": "YY",
                "IY": "ZY",
                "YI": "YZ",
                "YY": "XX",
                "IZ": "IZ",
                "ZI": "ZI",
                "ZZ": "ZZ",
                "XY": "-YX",
                "YX": "-XY",
                "XZ": "XI",
                "ZX": "IX",
                "YZ": "YI",
                "ZY": "IY",
            },
            cirq.ops.SWAP: {
                "II": "II",
                "IX": "XI",
                "XI": "IX",
                "XX": "XX",
                "IY": "YI",
                "YI": "IY",
                "YY": "YY",
                "IZ": "ZI",
                "ZI": "IZ",
                "ZZ": "ZZ",
                "XY": "YX",
                "YX": "XY",
                "XZ": "ZX",
                "ZX": "XZ",
                "YZ": "ZY",
                "ZY": "YZ",
            },
        }

        measurement_circuit = self._get_measurement_circuit().get_circuit()

        expect_val = 0.0
        for mermin_coef, mermin_pauli in self.mermin_operator:
            # Iterate through the operations in the measurement circuit and conjugate with the
            # current Pauli to determine the correct measurement qubits and coefficient.
            measure_pauli = [p for p in mermin_pauli]
            parity = 1
            for op in measurement_circuit.all_operations():
                if isinstance(op.gate, cirq.ops.MeasurementGate):
                    break

                substr = [measure_pauli[cast("cirq.LineQubit", qubit).x] for qubit in op.qubits]
                conjugated_substr = conjugation_rules[op.gate]["".join(substr)]

                if conjugated_substr[0] == "-":
                    parity = -1 * parity
                    conjugated_substr = conjugated_substr[1:]

                for qubit, pauli in zip(op.qubits, conjugated_substr):
                    measure_pauli[cast("cirq.LineQubit", qubit).x] = pauli

            measurement_qubits = [i for i, pauli in enumerate(measure_pauli) if pauli == "Z"]
            measurement_coef = parity

            numerator = 0.0
            for bitstr, count in counts.items():
                parity = 1
                for qb in measurement_qubits:
                    if bitstr[qb] == "1":  # Qubit order is big endian
                        parity = -1 * parity

                numerator += mermin_coef * measurement_coef * parity * count

            expect_val += numerator / sum(list(counts.values()))

        return (expect_val + 2 ** (self.num_qubits - 1)) / 2**self.num_qubits

    def _mermin_operator(self, num_qubits: int) -> list[tuple[float, str]]:
        """Generate the Mermin operator
        (https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.65.1838), or M_n
        (Eq. 2.8) in https://arxiv.org/pdf/2005.11271.pdf.
        """
        mermin_op = []
        for num_y in range(1, num_qubits + 1, 2):
            coef = (-1.0) ** (num_y // 2)

            for x_indices in itertools.combinations(range(num_qubits), num_qubits - num_y):
                pauli = np.array(["Y"] * num_qubits)
                pauli.put(x_indices, "X")
                mermin_op.append((coef, "".join(pauli)))

        return mermin_op

    def _get_measurement_circuit(self) -> stabilizers.MeasurementCircuit:
        """Return a MeasurementCircuit for simultaneous measurement of N operators.

        Each column of self.stabilizer represents a Pauli string that we seek to measure.
        Thus, self.stabilizer should have dimensions of 2 * N rows by N columns. The first N rows
        indicate the presence of a Z in each index of the Pauli String. The last N rows
        indicate X's.

        For instance, simultaneous measurement of YYI, XXY, IYZ would be represented by
        [[1, 0, 0],  ========
         [1, 0, 1],  Z matrix
         [0, 1, 1],  ========
         [1, 1, 0],  ========
         [1, 1, 1],  X matrix
         [0, 1, 0]   ========

        As annotated above, the submatrix of the first (last) N rows is referred to as
        the Z (X) matrix.

        All operators must commute and be independent (i.e. can't express any column as a base-2
        product of the other columns) for this code to work.
        """
        # Validate that the stabilizer matrix is valid
        assert self.stabilizer.shape == (
            2 * self.num_qubits,
            self.num_qubits,
        ), f"{self.num_qubits} qubits, but matrix shape: {self.stabilizer.shape}"

        # i, j will always denote row, column index
        for i in range(2 * self.num_qubits):
            for j in range(self.num_qubits):
                value = self.stabilizer[i, j]
                assert value in [0, 1], f"[{i}, {j}] index is {value}"

        measurement_circuit = stabilizers.MeasurementCircuit(
            cirq.Circuit(), self.stabilizer, self.num_qubits, self.qubits
        )

        stabilizers.prepare_X_matrix(measurement_circuit)
        stabilizers.row_reduce_X_matrix(measurement_circuit)
        stabilizers.patch_Z_matrix(measurement_circuit)
        stabilizers.change_X_to_Z_basis(measurement_circuit)
        # terminate with measurements
        measurement_circuit.get_circuit().append(cirq.measure(*self.qubits))

        return measurement_circuit



# Copyright 2021 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: disable-error-code=method-assign

from __future__ import annotations

import re
import uuid
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Any
from unittest.mock import MagicMock, patch

import cirq
import cirq_superstaq as css
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample, qcvv_resolver


def test_qcvv_resolver() -> None:
    assert qcvv_resolver("bad_name") is None
    assert qcvv_resolver("supermarq.qcvv.Sample") == Sample
    assert qcvv_resolver("supermarq.qcvv.QCVVExperiment") == QCVVExperiment

    # Check for something that is not explicitly exported
    assert qcvv_resolver("supermarq.qcvv.base_experiment.qcvv_resolver") is None


@dataclass
class ExampleResults(QCVVResults):
    """Example results class for testing."""

    _example_final_result: float | None = None

    def _analyze(self) -> None:
        self._example_final_result = 3.142

    def plot_results(self, filename: str | None = None) -> plt.Figure:
        fig = plt.Figure()
        if filename:
            fig.savefig(filename)
        return fig

    def print_results(self) -> None:
        print("This is a test")  # noqa: T201

    @property
    def example_final_result(self) -> float:
        if self._example_final_result is None:
            raise self._not_analyzed
        return self._example_final_result


class ExampleExperiment(QCVVExperiment[ExampleResults]):
    """Example experiment class for testing."""

    def __init__(
        self,
        qubits: int | Sequence[cirq.Qid],
        num_circuits: int,
        cycle_depths: Iterable[int],
        *,
        random_seed: int | None = None,
        _samples: list[Sample] | None = None,
        **kwargs: str | bool,
    ) -> None:
        super().__init__(
            qubits,
            num_circuits,
            cycle_depths,
            random_seed=random_seed,
            results_cls=ExampleResults,
            _samples=_samples,
            **kwargs,
        )

    def _build_circuits(self, num_circuits: int, cycle_depths: Iterable[int]) -> Sequence[Sample]:
        return [
            Sample(
                circuit=cirq.Circuit(cirq.measure(*self.qubits)),
                data={"depth": d},
                circuit_realization=k,
            )
            for k in range(num_circuits)
            for d in cycle_depths
        ]

    def _json_dict_(self) -> dict[str, Any]:
        return super()._json_dict_()


@pytest.fixture
def abc_experiment() -> ExampleExperiment:
    return ExampleExperiment(
        qubits=2,
        num_circuits=10,
        cycle_depths=[1, 3, 5],
        random_seed=42,
        service_details="Some other details",
    )


@pytest.fixture
def sample_circuits() -> list[Sample]:
    qubits = cirq.LineQubit.range(2)
    return [
        Sample(
            circuit=cirq.Circuit(
                cirq.X(qubits[1]), cirq.CZ(*qubits), cirq.CZ(*qubits), cirq.measure(*qubits)
            ),
            data={"circuit": 1},
            circuit_realization=1,
        ),
        Sample(
            circuit=cirq.Circuit(cirq.X(qubits[0]), cirq.CX(*qubits), cirq.measure(*qubits)),
            data={"circuit": 2},
            circuit_realization=2,
        ),
    ]


def test_qcvv_experiment_init(
    abc_experiment: ExampleExperiment,
) -> None:
    assert abc_experiment.num_qubits == 2
    assert abc_experiment.qubits == (cirq.q(0), cirq.q(1))
    assert abc_experiment.num_circuits == 10
    assert abc_experiment.cycle_depths == [1, 3, 5]
    assert abc_experiment._results_cls == ExampleResults
    assert abc_experiment._service_kwargs == {"service_details": "Some other details"}
    assert len(abc_experiment.samples) == 30
    assert isinstance(abc_experiment._rng, np.random.Generator)
    assert abc_experiment.circuits == [sample.circuit for sample in abc_experiment]

    new_experiment = ExampleExperiment(
        qubits=[cirq.q(1), cirq.q(3), cirq.q(7)],
        num_circuits=10,
        cycle_depths=[1, 3, 5],
    )
    assert new_experiment.num_qubits == 3
    assert new_experiment.qubits == (cirq.q(1), cirq.q(3), cirq.q(7))
    assert new_experiment.circuits == [sample.circuit for sample in new_experiment]


def test_results_init(
    abc_experiment: ExampleExperiment,
) -> None:
    results = ExampleResults(target="target", experiment=abc_experiment)
    assert results.target == "target"
    assert results.samples == abc_experiment.samples
    assert results.num_circuits == 10
    assert results.num_qubits == 2
    assert results.qubits == (cirq.q(0), cirq.q(1))


def test_results_getitem(
    abc_experiment: ExampleExperiment,
) -> None:
    q0, q1 = abc_experiment.qubits

    results = ExampleResults(
        target="example",
        experiment=abc_experiment,
        data=pd.DataFrame(
            [
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "00": 0.5,
                    "01": 0.5,
                    "10": 0.0,
                    "11": 0.0,
                }
                for sample in abc_experiment.samples
            ]
        ),
    )

    results_q0 = results[q0]
    assert results_q0.qubits == (q0,)
    assert results_q0.parent is results
    pd.testing.assert_frame_equal(
        results_q0.data,
        pd.DataFrame(
            [
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "0": 1.0,
                    "1": 0.0,
                }
                for sample in abc_experiment.samples
            ]
        ),
        check_like=True,
    )

    results_q1 = results[q1]
    assert results_q1.qubits == (q1,)
    assert results_q1.parent is results
    pd.testing.assert_frame_equal(
        results_q1.data,
        pd.DataFrame(
            [
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "0": 0.5,
                    "1": 0.5,
                }
                for sample in abc_experiment.samples
            ]
        ),
        check_like=True,
    )

    results_q0q1 = results[q0, q1]
    assert results_q0q1.qubits == (q0, q1)
    assert results_q0q1.parent is results
    pd.testing.assert_frame_equal(results_q0q1.data, results.data)

    results_q1q0 = results[q1, q0]
    assert results_q1q0.qubits == (q1, q0)
    assert results_q1q0.parent is results
    pd.testing.assert_frame_equal(
        results_q1q0.data,
        pd.DataFrame(
            [
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "00": 0.5,
                    "01": 0.0,
                    "10": 0.5,
                    "11": 0.0,
                }
                for sample in abc_experiment.samples
            ]
        ),
        check_like=True,
    )

    mock_job = MagicMock(spec=css.Job)
    mock_job.status.return_value = "Queued"
    results = ExampleResults(target="target", experiment=abc_experiment, job=mock_job)
    with pytest.raises(ValueError, match=r"No results to split."):
        _ = results[q0]


def test_experiment_init_with_bad_layers() -> None:
    with pytest.raises(
        ValueError, match=r"The `cycle_depths` iterator can only include positive values."
    ):
        ExampleExperiment(
            qubits=2,
            num_circuits=10,
            cycle_depths=[0],
            random_seed=42,
            service_details="Some other details",
        )


def test_results_not_analyzed(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(target="target", experiment=abc_experiment)
    with pytest.raises(
        RuntimeError,
        match=re.escape("Value has not yet been estimated. Please run `.analyze()` method."),
    ):
        _ = results.example_final_result


def test_results_job_still_running(abc_experiment: ExampleExperiment) -> None:
    mock_job = MagicMock(spec=css.Job)
    mock_job.status.return_value = "Pending"
    results = ExampleResults(target="target", experiment=abc_experiment, job=mock_job)
    with pytest.warns(
        Warning,
        match=(
            "Experiment data is not yet ready to analyse. This is likely because "
            "the Superstaq job has not yet been completed. Either wait and try again "
            "later, or interrogate the `.job` attribute."
        ),
    ):
        results.analyze()


def test_results_job_no_data(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(
        target="target",
        experiment=abc_experiment,
    )
    with pytest.raises(
        RuntimeError,
        match=(
            r"No data available and no Superstaq job to use to collect data. Please manually add "
            "results data in order to perform analysis"
        ),
    ):
        results.analyze()


def test_results_analyze(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(target="target", experiment=abc_experiment, data=pd.DataFrame())

    with (
        patch("matplotlib.pyplot.Figure.savefig") as mock_plot,
        patch("builtins.print") as mock_print,
    ):
        results.analyze(plot_results=True, print_results=True, plot_filename="test_name")

    assert results.example_final_result == 3.142
    mock_plot.assert_called_once_with("test_name")
    mock_print.assert_called_once_with("This is a test")


def test_results_ready(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(target="target", experiment=abc_experiment, data=pd.DataFrame())
    assert results.data_ready


def test_results_ready_from_job(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    mock_job = MagicMock(spec=css.Job)
    mock_job.status.return_value = "Done"
    mock_job.counts.return_value = [
        {
            "00": 20,
            "01": 5,
            "11": 10,
        },
        {
            "00": 30,
            "01": 5,
        },
    ]
    results = ExampleResults(target="target", experiment=abc_experiment, job=mock_job)
    assert results.data_ready
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {
                    "uuid": abc_experiment.samples[0].uuid,
                    "circuit_realization": 1,
                    "circuit": 1,
                    "00": 20 / 35,
                    "01": 5 / 35,
                    "10": 0.0,
                    "11": 10 / 35,
                },
                {
                    "uuid": abc_experiment.samples[1].uuid,
                    "circuit_realization": 2,
                    "circuit": 2,
                    "00": 30 / 35,
                    "01": 5 / 35,
                    "10": 0.0,
                    "11": 0.0,
                },
            ]
        ),
        check_like=True,
    )


def test_run_with_simulator(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits

    simulator = cirq.DensityMatrixSimulator()
    with patch("cirq.Simulator") as mock_sim:  # Mock default simulator to make sure it isn't called
        results = abc_experiment.run_with_simulator(simulator=simulator, repetitions=100)
        mock_sim.assert_not_called()

    assert results.experiment == abc_experiment
    assert results.target == "local_simulator"

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {
                    "uuid": abc_experiment.samples[0].uuid,
                    "circuit_realization": 1,
                    "circuit": 1,
                    "00": 0.0,
                    "01": 1.0,
                    "10": 0.0,
                    "11": 0.0,
                },
                {
                    "uuid": abc_experiment.samples[1].uuid,
                    "circuit_realization": 2,
                    "circuit": 2,
                    "00": 0.0,
                    "01": 0.0,
                    "10": 0.0,
                    "11": 1.0,
                },
            ]
        ),
    )


def test_run_with_simulator_default_target(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits

    results = abc_experiment.run_with_simulator(repetitions=100)

    assert results.experiment == abc_experiment
    assert results.target == "local_simulator"

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {
                    "uuid": abc_experiment.samples[0].uuid,
                    "circuit_realization": 1,
                    "circuit": 1,
                    "00": 0.0,
                    "01": 1.0,
                    "10": 0.0,
                    "11": 0.0,
                },
                {
                    "uuid": abc_experiment.samples[1].uuid,
                    "circuit_realization": 2,
                    "circuit": 2,
                    "00": 0.0,
                    "01": 0.0,
                    "10": 0.0,
                    "11": 1.0,
                },
            ]
        ),
    )


def test_run_on_device(abc_experiment: ExampleExperiment, sample_circuits: list[Sample]) -> None:
    abc_experiment.samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        results = abc_experiment.run_on_device(
            target="example_target", repetitions=100, some="options"
        )

    mock_service.return_value.create_job.assert_called_once_with(
        [sample_circuits[0].circuit, sample_circuits[1].circuit],
        target="example_target",
        method=None,
        repetitions=100,
        some="options",
    )

    assert results.job == mock_service.return_value.create_job.return_value
    assert results.target == "example_target"
    assert results.experiment == abc_experiment


def test_run_on_device_dry_run(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        results = abc_experiment.run_on_device(
            target="example_target", repetitions=100, method="dry-run"
        )

    mock_service.return_value.create_job.assert_called_once_with(
        [sample_circuits[0].circuit, sample_circuits[1].circuit],
        target="example_target",
        method="dry-run",
        repetitions=100,
    )
    assert results.job == mock_service.return_value.create_job.return_value
    assert results.target == "example_target"
    assert results.experiment == abc_experiment


def test_interleave_circuit(abc_experiment: ExampleExperiment) -> None:
    qubits = abc_experiment.qubits
    circuit = cirq.Circuit(*[cirq.X(qubits[0]) for _ in range(4)])

    # With last gate
    interleaved_circuit = abc_experiment._interleave_layer(
        circuit, cirq.Z(qubits[0]), include_final=True
    )
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
        ),
    )

    # Without last gate
    interleaved_circuit = abc_experiment._interleave_layer(
        circuit, cirq.Z(qubits[0]), include_final=False
    )
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
        ),
    )

    # Multi-gate layer
    layer = cirq.Moment(cirq.Z(qubits[0]), cirq.H(qubits[1]))
    interleaved_circuit = abc_experiment._interleave_layer(
        circuit, layer=layer, include_final=False
    )
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            cirq.H(qubits[1]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            cirq.H(qubits[1]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.Z(qubits[0]).with_tags("no_compile"),
            cirq.H(qubits[1]).with_tags("no_compile"),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
        ),
    )

    # Empty layer
    interleaved_circuit = abc_experiment._interleave_layer(circuit, layer=None, include_final=False)
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
            css.barrier(*qubits),
            cirq.X(qubits[0]),
        ),
    )


def test_validate_circuits(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    # Should't get any errors with the base circuits
    abc_experiment._validate_circuits(sample_circuits)

    # Add a gate so not all measurements are terminal
    sample_circuits[0].circuit += cirq.X(abc_experiment.qubits[0])
    with pytest.raises(
        ValueError, match=r"QCVV experiment circuits can only contain terminal measurements."
    ):
        abc_experiment._validate_circuits(sample_circuits)

    # Remove measurements
    sample_circuits[0].circuit = sample_circuits[0].circuit[:-2] + cirq.measure(
        abc_experiment.qubits[0]
    )
    with pytest.raises(
        ValueError,
        match=r"The terminal measurement in QCVV experiment circuits must measure all qubits.",
    ):
        abc_experiment._validate_circuits(sample_circuits)

    # Remove all measurements
    sample_circuits[0].circuit = sample_circuits[0].circuit[:-2]
    with pytest.raises(
        ValueError,
        match=r"QCVV experiment circuits must contain measurements.",
    ):
        abc_experiment._validate_circuits(sample_circuits)


def test_run_with_callable(abc_experiment: ExampleExperiment) -> None:
    def _example_callable(sample: Sample, some: str) -> dict[str, float]:
        assert sample
        assert some == "kwargs"
        return {"01": 0.2, "10": 0.7, "11": 0.1}

    results = abc_experiment.run_with_callable(
        _example_callable,  # type: ignore[arg-type]
        some="kwargs",
    )

    assert results.target == "callable"
    assert results.experiment == abc_experiment

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "00": 0.0,
                    "01": 0.2,
                    "10": 0.7,
                    "11": 0.1,
                }
                for sample in abc_experiment.samples
            ]
        ),
        check_like=True,
    )


def test_run_with_callable_mixed_keys(abc_experiment: ExampleExperiment) -> None:
    def _example_callable(sample: Sample, some: str) -> dict[str | int, float]:
        assert sample
        assert some == "kwargs"
        return {1: 0.2, "10": 0.7, 3: 0.1}

    results = abc_experiment.run_with_callable(
        _example_callable,  # type: ignore[arg-type]
        some="kwargs",
    )

    assert results.target == "callable"
    assert results.experiment == abc_experiment

    # Check the data is stored
    pd.testing.assert_frame_equal(
        results.data,
        pd.DataFrame(
            [
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "00": 0.0,
                    "01": 0.2,
                    "10": 0.7,
                    "11": 0.1,
                }
                for sample in abc_experiment.samples
            ]
        ),
        check_like=True,
    )


def test_run_with_callable_bad_bitstring(abc_experiment: ExampleExperiment) -> None:
    def _example_callable(sample: Sample, some: str) -> dict[str, float]:
        assert sample
        assert some == "kwargs"
        return {"000": 0.0, "01": 0.2, "10": 0.8}

    with pytest.raises(
        ValueError,
        match=(r"The key contains the wrong number of bits. Got 3 entries but expected 2 bits."),
    ):
        abc_experiment.run_with_callable(_example_callable, some="kwargs")  # type: ignore[arg-type]


def test_results_collect_device_counts(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits

    mock_job = MagicMock(spec=css.Job)
    mock_job.counts.return_value = [
        {
            "00": 20,
            "01": 5,
            "11": 10,
        },
        {
            "00": 30,
            "01": 5,
        },
    ]
    results = ExampleResults(target="example_target", experiment=abc_experiment, job=mock_job)

    df = results._collect_device_counts()

    pd.testing.assert_frame_equal(
        df,
        pd.DataFrame(
            [
                {
                    "uuid": abc_experiment.samples[0].uuid,
                    "circuit_realization": 1,
                    "circuit": 1,
                    "00": 20 / 35,
                    "01": 5 / 35,
                    "10": 0.0,
                    "11": 10 / 35,
                },
                {
                    "uuid": abc_experiment.samples[1].uuid,
                    "circuit_realization": 2,
                    "circuit": 2,
                    "00": 30 / 35,
                    "01": 5 / 35,
                    "10": 0.0,
                    "11": 0.0,
                },
            ]
        ),
        check_like=True,
    )


def test_results_collect_device_counts_no_job(abc_experiment: ExampleExperiment) -> None:
    results = ExampleResults(target="example_target", experiment=abc_experiment, job=None)
    with pytest.raises(
        ValueError,
        match=(r"No Superstaq job associated with these results. Cannot collect device counts."),
    ):
        results._collect_device_counts()


def test_results_from_records(abc_experiment: ExampleExperiment) -> None:
    # All accepted types
    records_1 = {s.uuid: {"01": 1, "10": 3} for s in abc_experiment.samples}
    records_2 = {s.uuid: {"01": 0.25, "10": 0.75} for s in abc_experiment.samples}
    records_3 = {s.uuid: {1: 1, 2: 3} for s in abc_experiment.samples}
    records_4 = {s.uuid: {1: 0.25, 2: 0.75} for s in abc_experiment.samples}

    for record in (records_1, records_2, records_3, records_4):
        results = abc_experiment.results_from_records(record)
        pd.testing.assert_frame_equal(
            results.data,
            pd.DataFrame(
                {
                    "uuid": sample.uuid,
                    "circuit_realization": sample.circuit_realization,
                    "depth": sample.data["depth"],
                    "00": 0.0,
                    "01": 0.25,
                    "10": 0.75,
                    "11": 0.0,
                }
                for sample in abc_experiment.samples
            ),
        )


def test_results_from_records_bad_input(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    # Warn for missing samples
    with pytest.warns(
        UserWarning,
        match=re.escape(
            f"The following samples are missing records: {sample_circuits[1].uuid!s}. These "
            "will not be included in the results."
        ),
    ):
        abc_experiment.results_from_records({sample_circuits[0].uuid: {"00": 10}})

    # Warn for spurious records
    new_uuid = uuid.uuid4()
    with pytest.warns(
        UserWarning,
        match=re.escape("Unable to find matching sample for 1 record(s)."),
    ):
        abc_experiment.results_from_records(
            {
                new_uuid: {"00": 10},
                sample_circuits[0].uuid: {"00": 10},
                sample_circuits[1].uuid: {"00": 10},
            }
        )

    # Error when processing samples
    with pytest.raises(ValueError, match=re.escape("No non-zero counts.")):
        abc_experiment.results_from_records(
            {
                sample_circuits[0].uuid: {"00": 0},
                sample_circuits[1].uuid: {"00": 0},
            }
        )


def test_canonicalize_bitstring() -> None:
    assert QCVVExperiment.canonicalize_bitstring("00", 2) == "00"
    assert QCVVExperiment.canonicalize_bitstring(1, 2) == "01"
    assert QCVVExperiment.canonicalize_bitstring(5, 4) == "0101"

    with pytest.raises(ValueError, match=r"The key must be positive. Instead got -2."):
        QCVVExperiment.canonicalize_bitstring(-2, 4)

    with pytest.raises(
        ValueError,
        match=(
            r"The key is too large to be encoded with 4 qubits. Got 72 but expected less than 16."
        ),
    ):
        QCVVExperiment.canonicalize_bitstring(72, 4)

    with pytest.raises(
        ValueError,
        match=(r"The key contains the wrong number of bits. Got 5 entries but expected 4 bits."),
    ):
        QCVVExperiment.canonicalize_bitstring("01010", 4)

    with pytest.raises(ValueError, match=r"All entries in the bitstring must be 0 or 1. Got 1234."):
        QCVVExperiment.canonicalize_bitstring("1234", 4)

    with pytest.raises(TypeError, match=r"Key must either be `numbers.Integral` or `str`."):
        QCVVExperiment.canonicalize_bitstring(3.141, 4)  # type: ignore[arg-type]


def test_canonicalize_probabilities() -> None:
    p1 = {0: 0.1, 1: 0.6, 3: 0.3}
    p2 = {0: 1, 1: 6, 3: 3}
    p3 = {"00": 0.1, "01": 0.6, "11": 0.3}
    p4 = {"00": 1, "01": 6, "11": 3}
    p_list: list[dict[str, int] | dict[str, float] | dict[int, int] | dict[int, float]] = [
        p1,
        p2,
        p3,
        p4,
    ]
    for p in p_list:
        assert QCVVExperiment.canonicalize_probabilities(p, 2) == {
            "00": 0.1,
            "01": 0.6,
            "10": 0.0,
            "11": 0.3,
        }

    # Test for empty dictionary
    assert QCVVExperiment.canonicalize_probabilities({}, 2) == {}


def test_canonicalize_probabilities_bad_input() -> None:
    # Negative counts
    with pytest.raises(ValueError, match=r"Probabilities/counts must be positive."):
        QCVVExperiment.canonicalize_probabilities({0: -2}, 2)

    # No non-zero counts
    with pytest.raises(ValueError, match=r"No non-zero counts."):
        QCVVExperiment.canonicalize_probabilities({0: 0, 1: 0}, 2)

    # Negative probabilities
    with pytest.raises(ValueError, match=r"Probabilities/counts must be positive."):
        QCVVExperiment.canonicalize_probabilities({0: 0.0, 1: -0.5}, 2)


def test_experiment_get_item(
    abc_experiment: ExampleExperiment,
    sample_circuits: list[Sample],
) -> None:
    abc_experiment.samples = sample_circuits

    for k, _ in enumerate(sample_circuits):
        assert abc_experiment[k] == sample_circuits[k]
        assert abc_experiment[sample_circuits[k].uuid] == sample_circuits[k]
        assert abc_experiment[str(sample_circuits[k].uuid)] == sample_circuits[k]

    with pytest.raises(TypeError, match=r"Key must be int, str or uuid.UUID"):
        _ = abc_experiment[3.141]  # type: ignore[index]

    with pytest.raises(
        KeyError, match=re.escape("No sample found with UUID b55adabc-39c4-4f7b-a84d-906adaf0897e")
    ):
        _ = abc_experiment["b55adabc-39c4-4f7b-a84d-906adaf0897e"]

    with pytest.raises(
        RuntimeError, match=r"Multiple samples found with matching key. Something has gone wrong."
    ):
        # Manually set duplicate sample uuids
        abc_experiment.samples[0].uuid = abc_experiment.samples[1].uuid
        _ = abc_experiment[sample_circuits[0].uuid]


def test_map_records_to_samples(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits

    records = {
        0: {0: 0.1, 1: 0.6, 3: 0.3},
        sample_circuits[1].uuid: {0: 4, 1: 6, 3: 2},
    }
    mapped_samples = abc_experiment._map_records_to_samples(records)  # type: ignore[arg-type]
    assert mapped_samples == {
        sample_circuits[0]: {0: 0.1, 1: 0.6, 3: 0.3},
        sample_circuits[1]: {0: 4, 1: 6, 3: 2},
    }


def test_map_records_to_samples_missing_key(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    with (
        pytest.warns(
            UserWarning, match=re.escape("Unable to find matching sample for 1 record(s).")
        ),
        pytest.warns(
            UserWarning,
            match=(
                f"The following samples are missing records: {sample_circuits[0].uuid}. "
                "These will not be included in the results."
            ),
        ),
    ):
        abc_experiment._map_records_to_samples(
            {
                5: {0: 0.1, 1: 0.6, 3: 0.3},
                sample_circuits[1].uuid: {0: 4, 1: 6, 3: 2},
            }
        )


def test_map_records_to_samples_bad_key_type(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    # Bad key type
    with pytest.raises(
        TypeError,
        match=re.escape("Key must be int, str or uuid.UUID, not <class 'float'>"),
    ):
        abc_experiment._map_records_to_samples(
            {
                5.0: {0: 0.1, 1: 0.6, 3: 0.3},  # type: ignore[dict-item]
                sample_circuits[1].uuid: {0: 4, 1: 6, 3: 2},
            }
        )


def test_map_records_to_samples_duplicate_keys(
    abc_experiment: ExampleExperiment, sample_circuits: list[Sample]
) -> None:
    abc_experiment.samples = sample_circuits
    with pytest.raises(
        KeyError,
        match=re.escape(
            f"Duplicate records found for sample with uuid: {sample_circuits[1].uuid!s}."
        ),
    ):
        abc_experiment._map_records_to_samples(
            {
                1: {0: 0.1, 1: 0.6, 3: 0.3},
                sample_circuits[1].uuid: {0: 4, 1: 6, 3: 2},
            }
        )


@patch("supermarq.qcvv.base_experiment.qcvv_resolver")
def test_dump_and_load(
    mock_resolver: MagicMock,
    tmp_path_factory: pytest.TempPathFactory,
    abc_experiment: ExampleExperiment,
    sample_circuits: list[Sample],
) -> None:
    temp_resolver = {
        "supermarq.qcvv.Sample": Sample,
        "supermarq.qcvv.ExampleExperiment": ExampleExperiment,
    }
    mock_resolver.side_effect = lambda x: temp_resolver.get(x, qcvv_resolver(x))

    filename = tmp_path_factory.mktemp("tempdir") / "file.json"
    abc_experiment.samples = sample_circuits
    abc_experiment.to_file(filename)
    exp = ExampleExperiment.from_file(filename)

    assert exp.samples == abc_experiment.samples
    assert exp.num_qubits == abc_experiment.num_qubits
    assert exp.num_circuits == abc_experiment.num_circuits
    assert exp.cycle_depths == abc_experiment.cycle_depths


def test_count_non_barrier_gates() -> None:
    qubits = cirq.LineQubit.range(4)
    circuit = cirq.Circuit(
        cirq.X(qubits[0]),
        cirq.Y(qubits[1]),
        css.barrier(*qubits),
        cirq.CX(*qubits[0:2]),
        css.barrier(*qubits),
        css.ParallelRGate(0, 0, num_copies=4).on(*qubits),
    )
    assert ExampleExperiment._count_non_barrier_gates(circuit) == 4
    assert ExampleExperiment._count_non_barrier_gates(circuit, num_qubits=1) == 2
    assert ExampleExperiment._count_non_barrier_gates(circuit, num_qubits=2) == 1
    assert ExampleExperiment._count_non_barrier_gates(circuit, num_qubits=4) == 1



from __future__ import annotations

import cirq

import supermarq
from supermarq.benchmarks.qaoa_fermionic_swap_proxy import QAOAFermionicSwapProxy


def test_qaoa_circuit() -> None:
    """Test the circuit generation function."""
    qaoa = QAOAFermionicSwapProxy(4)
    assert len(qaoa.circuit().all_qubits()) == 4
    assert (
        len(
            list(qaoa.circuit().findall_operations(lambda op: isinstance(op.gate, type(cirq.CNOT))))
        )
        == 18
    )


def test_qaoa_score() -> None:
    """Test the score evaluation function."""
    qaoa = QAOAFermionicSwapProxy(4)
    # Reverse bitstring ordering due to SWAP network
    raw_counts = supermarq.simulation.get_ideal_counts(qaoa.circuit())
    ideal_counts = {bitstring[::-1]: probability for bitstring, probability in raw_counts.items()}
    assert qaoa.score({k[::-1]: v for k, v in ideal_counts.items()}) > 0.99



from __future__ import annotations

from collections.abc import Mapping

import cirq
import pytest
from qiskit.quantum_info import hellinger_fidelity

from supermarq.benchmark import Benchmark


@pytest.fixture
def benchmark() -> Benchmark:
    """Simple one-qubit benchmark that creates an equal superposition state.

    Returns:
        A benchmark instance.
    """

    class _TestBenchmark(Benchmark):
        def circuit(self) -> cirq.Circuit:
            """Returns:
            A test cirq circuit of an equal superposition state.
            """
            qubit = cirq.LineQubit(0)
            return cirq.Circuit(cirq.H(qubit), cirq.measure(qubit))

        def score(self, counts: Mapping[str, float]) -> float:
            """Device performace score.

            Args:
                counts: Dictionary(s) containing the measurement counts from execution.

            Returns:
                a normalized [0,1] score reflecting device performance.
            """
            dist = {b: c / sum(counts.values()) for b, c in counts.items()}
            return hellinger_fidelity({"0": 0.5, "1": 0.5}, dist)

    return _TestBenchmark()


def test_benchmark_circuit(benchmark: Benchmark) -> None:
    circuit = cirq.Circuit(cirq.H(cirq.LineQubit(0)), cirq.measure(cirq.LineQubit(0)))
    assert benchmark.circuit() == circuit


def test_benchmark_score(benchmark: Benchmark) -> None:
    assert benchmark.score({"0": 50, "1": 50}) == 1



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Circle
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from sklearn.linear_model import LinearRegression

if TYPE_CHECKING:
    import numpy.typing as npt


def plot_results(
    scores: list[float], tick_labels: list[str], savefn: str | None = None, show: bool = True
) -> None:
    """Plot a simple bar chart of the benchmark results.

    The ordering of scores and labels is assumed to be the same.

    Args:
        scores: List of benchmark results.
        tick_labels: List of quantum device names.
        savefn: Path to save the plot, if `None`, the plot is not saved.
        show: Display the plot using `plt.show`.
    """
    _, ax = plt.subplots(dpi=150)

    width = 0.4

    xvals = np.arange(len(scores))

    ax.bar(xvals, scores, width=width, tick_label=tick_labels, align="center")

    ax.set_xticks(range(len(scores)))
    ax.set_ylabel("Score")
    plt.tight_layout()

    if savefn is not None:
        # Don't want to save figures when running tests
        plt.savefig(savefn)  # pragma: no cover

    if show:
        # Tests will hang if we show figures during tests
        plt.show()  # pragma: no cover

    plt.close()


def plot_volumetric_results(
    benchmark_data: Iterable[tuple[int, int, float]],
    savefn: str | None = None,
    title: str | None = None,
    show: bool = True,
    xmax: float = 60,
    ymax: float = 7,
    rect_width: float = 1.5,
    rect_height: float = 0.2,
) -> None:
    """Plot the benchmark results on an (x, y) = (depth, qubits) axis.

    It is assumed that all of the given data was collected on a single device. Keyword arguments
    may need to be tweaked depending on the domain and range of benchmark data.

    Args:
        benchmark_data: List of tuples containing (circuit depth, qubits, score) for each benchmark
            executed on a single device.
        savefn: Path to save the plot, if `None`, the plot is not saved.
        title: Optional title for the plot.
        show: Display the plot using `plt.show`.
        xmax: The rightmost limit of the x axis.
        ymax: The uppermost limit of the y axis.
        rect_width: The width of the rectangles.
        rect_height: The height of the rectangles.
    """
    _, ax = plt.subplots(figsize=[4, 4])

    cmap = mpl.colormaps["RdBu"]
    norm = mpl.colors.Normalize(vmin=0, vmax=1)

    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    plt.colorbar(sm, ax=ax, label="Benchmark Score")

    for x, y, z in benchmark_data:
        rect = plt.Rectangle(
            (x - rect_width / 2, y - rect_height / 2),
            rect_width,
            rect_height,
            color=cmap(norm(z)),
        )
        ax.add_patch(rect)

    ax.set_xlim(0, xmax)
    ax.set_ylim(0, ymax)
    ax.set_xlabel("Depth")
    ax.set_ylabel("Qubits")
    if title:
        ax.set_title(title)

    if savefn:
        # Don't want to save figures when running tests
        plt.savefig(savefn, bbox_inches="tight")  # pragma: no cover

    if show:
        # Tests will hang if we show figures during tests
        plt.show()  # pragma: no cover
    plt.close()


def plot_correlations(
    benchmark_features: dict[str, list[float]],
    device_scores: dict[str, float] | Iterable[dict[str, float]],
    feature_labels: list[str],
    device_name: str | list[str],
    savefn: str | None = None,
    show: bool = True,
) -> None:
    """Plot a correlation heatmap of the features for one or multiple devices.

    Args:
        benchmark_features: A dictionary where the keys are benchmark names and the values are the
            list of feature values for that benchmark.
        device_scores: A dictionary of (benchmark name, score) pairs, or a list of such
            dictionaries.
        feature_labels: Feature names, should have the same length as the lists of feature values
            in `benchmark_features`.
        device_name: The name or list of names of quantum device(s) where the scores were obtained.
        savefn: Path to save the plot, if `None`, the plot is not saved.
        show: Display the plot using `plt.show`.
    """
    temp_correlations = []
    if isinstance(device_scores, dict):
        device_scores = [device_scores]
    for cur_device_scores in device_scores:
        device_correlations = []
        for i in range(len(feature_labels)):
            x, y = [], []
            for benchmark in cur_device_scores:
                x.append(benchmark_features[benchmark][i])
                y.append(cur_device_scores[benchmark])

            proper_x = np.array(x)[:, np.newaxis]
            proper_y = np.array(y)
            model = LinearRegression().fit(proper_x, proper_y)
            correlation = model.score(proper_x, proper_y)
            device_correlations.append(correlation)
        temp_correlations.append(device_correlations)

    correlations = np.array(temp_correlations)

    if isinstance(device_name, str):
        device_names = [device_name]
    else:
        device_names = device_name

    _, ax = plt.subplots(dpi=300)
    im, _ = heatmap(
        correlations,
        ax,
        device_names,
        feature_labels,
        cmap="cool",
        vmin=0,
        cbarlabel=r"Coefficient of Determination, $R^2$",
        cbar_kw={"pad": 0.005},
    )

    annotate_heatmap(im, size=7)

    if savefn is not None:
        # Don't want to save figures when running tests
        plt.savefig(savefn, bbox_inches="tight")  # pragma: no cover

    if show:
        # Tests will hang if we show figures during tests
        plt.show()  # pragma: no cover
    plt.close()


def plot_benchmark(
    title: str,
    labels: list[str],
    features: list[list[float]],
    show: bool = True,
    savefn: str | None = None,
    spoke_labels: list[str] | None = None,
    legend_loc: tuple[float, float] = (0.75, 0.85),
) -> None:
    """Create a radar plot showing the feature vectors of the given benchmarks.

    Args:
        title: The string title of the plot.
        labels: A list of string benchmark labels for the plot.
        features: A list of feature data in the format: [[features_1], [features_2], ...]].
        show: Boolean flag to display the plot using `plt.show`.
        savefn: Path to save the plot, if `None`, the plot is not saved.
        spoke_labels: Optional labels for the feature vector dimensions.
        legend_loc: Optional argument to fine tune the legend placement.
    """
    if spoke_labels is None:
        spoke_labels = ["Connectivity", "Liveness", "Parallelism", "Measurement", "Entanglement"]

    num_spokes = len(spoke_labels)
    theta = radar_factory(num_spokes)

    _, ax = plt.subplots(dpi=150, subplot_kw={"projection": "radar"})
    assert isinstance(ax, RadarAxesMeta)

    ax.set_rgrids([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])
    ax.set_title(
        str(title),
        weight="bold",
        size="medium",
        position=(0.5, 1.1),
        horizontalalignment="center",
        verticalalignment="center",
    )
    for d, label in zip(features, labels):
        ax.plot(theta, d, label=label)
        ax.fill(theta, d, alpha=0.25)
    ax.set_varlabels(spoke_labels)

    ax.legend(loc=legend_loc, labelspacing=0.1, fontsize=11)
    plt.tight_layout()

    if savefn is not None:
        # Don't want to save figures when running tests
        plt.savefig(savefn)  # pragma: no cover

    if show:
        # Tests will hang if we show figures during tests
        plt.show()  # pragma: no cover

    plt.close()


def heatmap(
    data: npt.NDArray[np.floating[Any]],
    ax: mpl.axes.Axes,
    row_labels: list[str],
    col_labels: list[str],
    cbar_kw: dict[str, Any] | None = None,
    cbarlabel: str = "",
    **kwargs: Any,
) -> tuple[mpl.image.AxesImage, Any]:
    """Create a heatmap from a numpy array and two lists of labels.

    Args:
        data: A 2D numpy array of shape (N, M).
        row_labels: A list or array of length N with the labels for the rows.
        col_labels: A list or array of length M with the labels for the columns.
        ax: An optional `matplotlib.axes.Axes` instance to which the heatmap is plotted.
            If not provided, use current axes or create a new one.
        cbar_kw: An optional dictionary with arguments to `matplotlib.Figure.colorbar`.
        cbarlabel: An optional label for the colorbar.
        kwargs: All other arguments are forwarded to `imshow`.

    Returns:
        The generated heatmap and the associated color bar.
    """
    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # Create colorbar
    if cbar_kw is None:
        cbar_kw = {}  # pragma: no cover

    figure = ax.get_figure()
    assert figure
    cbar = figure.colorbar(im, ax=ax, orientation="horizontal", **cbar_kw)
    cbar.ax.set_xlabel(cbarlabel, labelpad=10, va="bottom", fontsize=8)
    cbar.ax.tick_params(labelsize=10)

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels, fontsize=8)  # regular fontsize is 12
    ax.set_yticklabels(row_labels, fontsize=7)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=-30, ha="right", rotation_mode="anchor")

    # Turn spines off and create white grid.
    ax.spines[:].set_visible(False)

    ax.set_xticks(np.arange(data.shape[1] + 1) - 0.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - 0.5, minor=True)
    ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)

    # Rotate the 3 typical features separately
    for t in ax.get_xticklabels()[-3:]:
        t.set_horizontalalignment("left")
        t.set_rotation(35)

    return im, cbar


def annotate_heatmap(
    im: mpl.image.AxesImage,
    data: npt.NDArray[np.floating[Any]] | None = None,
    valfmt: Any = "{x:.2f}",
    textcolors: tuple[str, str] = ("black", "white"),
    threshold: float | None = None,
    **textkw: Any,
) -> list[mpl.text.Text]:
    """Annotate the given heatmap.

    Args:
        im: The `AxesImage` to be labeled.
        data: Optional data used to annotate. If None, the image's data is used.
        valfmt: An optional format of the annotations inside the heatmap.  This should either
            use the string format method, e.g. "$ {x:.2f}", or be a `matplotlib.ticker.Formatter`.
        textcolors: A pair of colors.  The first is used for values below a threshold, the second
            for those above. Defaults to black and white respectively.
        threshold: An optional value in data units according to which of the colors from textcolors
            are applied. If None (the default) uses the middle of the colormap as separation.
        textkw: All other arguments are forwarded to each call to `text` used to create
            the text labels.

    Returns:
        List of the text annotations.
    """
    if data is None:
        data = np.asarray(im.get_array())

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)  # pragma: no cover
    else:
        threshold = im.norm(data.max()) / 2.0

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw: dict[str, Any] = {"horizontalalignment": "center", "verticalalignment": "center"}
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = mpl.ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts


def radar_factory(num_vars: int) -> npt.NDArray[np.float64]:
    """Create a radar chart with `num_vars` axes.

    (https://matplotlib.org/stable/gallery/specialty_plots/radar_chart.html) This function
    creates a `RadarAxes` projection and registers it.

    Args:
        num_vars: Number of variables for radar chart.

    Returns:
        A list of evenly spaced angles.
    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False, dtype=np.float64)

    class RadarAxes(RadarAxesMeta):
        """A helper class that sets the shape of the feature plot."""

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            """Initializes the helper `RadarAxes` class."""
            self.frame = "circle"
            self.theta = theta
            self.num_vars = num_vars
            super().__init__(*args, **kwargs)

    register_projection(RadarAxes)
    return theta


class RadarAxesMeta(PolarAxes):
    """A helper class to generate feature-vector plots."""

    name = "radar"
    # use 1 line segment to connect specified points
    RESOLUTION = 1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the `RadarAxesMeta` class."""
        self.theta: npt.NDArray[np.float64]
        self.num_vars: int
        self.frame: str

        super().__init__(*args, **kwargs)
        # rotate plot such that the first axis is at the top
        self.set_theta_zero_location("N")

    def fill(self, *args: Any, closed: bool = True, **kwargs: Any) -> list[mpl.patches.Polygon]:
        """Method to override fill so that line is closed by default.

        Args:
            args: Arguments to be passed to fill.
            closed: Optional parameter to close fill or not. Defaults to True.
            kwargs: Other desired keyworded arguments to be passed to fill.

        Returns:
            A list of `matplotlib.patches.Polygon`.
        """
        return super().fill(*args, closed=closed, **kwargs)

    def plot(self, *args: Any, **kwargs: Any) -> list[mpl.lines.Line2D]:
        """Overrides plot so that line is closed by default.

        Args:
            args: Desired arguments for plotting.
            kwargs: Other desired keyword arguments for plotting.

        Returns:
            New closed lines.
        """
        lines = super().plot(*args, **kwargs)
        for line in lines:
            self._close_line(line)

        return lines

    def _close_line(self, line: mpl.lines.Line2D) -> None:
        """A method to close the input line.

        Args:
            line: The line to close.
        """
        x, y = map(np.asarray, line.get_data())
        # FIXME: markers at x[0], y[0] get doubled-up.
        # See issue https://github.com/Infleqtion/client-superstaq/issues/27
        if x[0] != x[-1]:
            x = np.append(x, x[0])
            y = np.append(y, y[0])
            line.set_data(x, y)

    def set_varlabels(self, labels: list[str]) -> None:
        """Sets the spoke labels at the appropriate points on the radar plot.

        Args:
            labels: The list of labels to apply.
        """
        self.set_thetagrids(np.degrees(self.theta), labels, fontsize=14)

    def _gen_axes_patch(self) -> mpl.patches.Circle:
        # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
        # in axes coordinates.
        return Circle((0.5, 0.5), 0.5)



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for SU(2) benchmarking."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import cirq
import cirq_superstaq as css
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.ticker import MaxNLocator
from scipy.stats import linregress
from tqdm.contrib.itertools import product

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample

if TYPE_CHECKING:
    from typing import Self


@dataclass
class SU2Results(QCVVResults):
    """Data structure for the SU2 experiment results."""

    _two_qubit_gate_fidelity: float | None = None
    """Estimated two qubit gate fidelity"""
    _two_qubit_gate_fidelity_std: float | None = None
    """Standard deviation of the two qubit gate fidelity estimate"""
    _single_qubit_noise: float | None = None
    """Estimated single qubit noise."""
    _single_qubit_noise_std: float | None = None
    """Standard deviation of the single qubit noise estimate"""

    @property
    def two_qubit_gate_fidelity(self) -> float:
        """Returns:
        Estimated two qubit gate fidelity."""
        if self._two_qubit_gate_fidelity is None:
            raise self._not_analyzed
        return self._two_qubit_gate_fidelity

    @property
    def two_qubit_gate_fidelity_std(self) -> float:
        """Returns:
        Standard deviation of estimated two qubit gate fidelity."""
        if self._two_qubit_gate_fidelity_std is None:
            raise self._not_analyzed
        return self._two_qubit_gate_fidelity_std

    @property
    def single_qubit_noise(self) -> float:
        """Returns:
        Estimated single qubit noise."""
        if self._single_qubit_noise is None:
            raise self._not_analyzed
        return self._single_qubit_noise

    @property
    def single_qubit_noise_std(self) -> float:
        """Returns:
        Standard deviation of estimated single qubit noise."""
        if self._single_qubit_noise_std is None:
            raise self._not_analyzed
        return self._single_qubit_noise_std

    @property
    def two_qubit_gate_error(self) -> float:
        """Returns:
        The two qubit gate error. Equal to one minus the fidelity.
        """
        return 1 - self.two_qubit_gate_fidelity

    @property
    def two_qubit_gate_error_std(self) -> float:
        """Returns:
        The two qubit gate error standard deviation. Equal to standard deviation of the
        fidelity.
        """
        return self.two_qubit_gate_fidelity_std

    def plot_results(self, filename: str | None = None) -> plt.Figure:
        """Plot the results of the experiment.

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.

        Returns:
            A single matplotlib figure containing the relevant plots of the results data.

        Raises:
            RuntimeError: If there is no data stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot plot results.")

        fig, ax = plt.subplots()
        sns.scatterplot(
            data=self.data.drop(columns=["circuit_realization", "uuid"]).melt(
                id_vars="num_two_qubit_gates", var_name="state", value_name="prob"
            ),
            x="num_two_qubit_gates",
            y="prob",
            hue="state",
            hue_order=["00", "01", "10", "11"],
            style="state",
            ax=ax,
        )
        ax.plot(
            xx := self.data["num_two_qubit_gates"],
            3 / 4 * (1 - self.single_qubit_noise) * self.two_qubit_gate_fidelity**xx + 0.25,
            label="00 (fit)",
        )
        ax.set_xlabel("Number of two qubit gates")
        ax.set_ylabel("State probability")
        ax.legend(title="State")
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))

        fig.tight_layout()

        if filename is not None:
            fig.savefig(filename)

        return fig

    def _analyze(self) -> None:
        """Perform the experiment analysis and store the results in the `results` attribute.

        Args:
            plot_results: Whether to generate plots of the results. Defaults to False.

        Returns:
            A named tuple of the final results from the experiment.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot perform analysis.")

        fit = linregress(
            x=self.data["num_two_qubit_gates"],
            y=np.log(4 / 3 * (self.data["00"] - 1 / 4)),
            # 1/4 < self.data["00"] < 1 so we subtract 1/4 and rescale by 4/3 to obtain a
            # quantity in the range 0 < 4 / 3 * (self.data["00"] - 1 / 4) < 1
        )
        gate_fid = np.exp(fit.slope)
        gate_fid_std = fit.stderr * gate_fid

        single_qubit_noise = 1 - np.exp(fit.intercept)
        single_qubit_noise_std = fit.intercept_stderr * (1 - single_qubit_noise)

        # Save results
        self._two_qubit_gate_fidelity = gate_fid
        self._two_qubit_gate_fidelity_std = gate_fid_std
        self._single_qubit_noise = single_qubit_noise
        self._single_qubit_noise_std = single_qubit_noise_std

    def print_results(self) -> None:
        """Prints the key results data."""
        print(  # noqa: T201
            f"Estimated two qubit gate fidelity: {self.two_qubit_gate_fidelity:.5} "
            f"+/- {self.two_qubit_gate_error_std:.5}\n"
            f"Estimated single qubit noise: {self.single_qubit_noise:.5} "
            f"+/- {self.single_qubit_noise_std:.5}\n"
        )


class SU2(QCVVExperiment[SU2Results]):
    r"""SU2 benchmarking experiment.

    SU2 benchmarking extracts the fidelity of a given two qubit gate, even in the presence of
    additional single qubit errors. The method works by sampling circuits of the form

    .. code::

        0: ──|─Rr───Q───X───Q──|─ ^{n} ... ─|─Rr───X─|─ ^{N-n} ... ──Rf───M───
             |      │       │  |            |        |                    │
        1: ──|─Rr───Q───X───Q──|─      ... ─|─Rr───X─|─        ... ──Rf───M───

    Where each :code:`Rr` gate is a randomly chosen :math:`SU(2)` rotation and the :code:`Rf` gates
    are single qubit :math:`SU(2)` rotations that in the absence of noise invert the preceding
    circuit so that the final qubit state should be :code:`00`.

    An exponential fit decay is then fitted to the observed 00 state probability as it decays with
    the number of two qubit gates included. Note that all circuits contain a fixed number of single
    qubit gates, so that the contribution for single qubit noise is constant.

    See Fig. 3 of https://www.nature.com/articles/s41586-023-06481-y#Fig3 for further
    details.
    """

    def __init__(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
        two_qubit_gate: cirq.Gate = cirq.CZ,
        *,
        random_seed: int | np.random.Generator | None = None,
        _samples: list[Sample] | None = None,
        **kwargs: str,
    ) -> None:
        """Args:
        two_qubit_gate: The Clifford gate to measure the gate error of.
        num_qubits: The number of qubits to experiment on. Must equal 2.
        """
        if two_qubit_gate.num_qubits() != 2:
            raise ValueError(
                "The `two_qubit_gate` parameter must be a gate that acts on exactly two qubits."
            )
        self.two_qubit_gate = two_qubit_gate
        """The two qubit gate to be benchmarked"""

        super().__init__(
            qubits=2,
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            random_seed=random_seed,
            results_cls=SU2Results,
            _samples=_samples,
            **kwargs,
        )

    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[Sample]:
        """Build a list of circuits required for the experiment.

        These circuits are stored in :class:`Sample` objects along with any additional data that is
        needed during the analysis.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different cycle depths to use during the experiment.

        Returns:
           The list of experiment samples.
        """
        samples = []
        max_depth = max(cycle_depths)
        for depth, index in product(cycle_depths, range(num_circuits), desc="Building circuits"):
            circuit = cirq.Circuit(
                *[self._component(include_two_qubit_gate=True) for _ in range(depth)],
                *[self._component(include_two_qubit_gate=False) for _ in range(max_depth - depth)],
            )
            circuit_inv = cirq.inverse(circuit)
            # Decompose circuit inverse into a pair of single qubit rotation gates
            _, rot_1, rot_2 = cirq.kron_factor_4x4_to_2x2s(cirq.unitary(circuit_inv))

            if (op_1 := cirq.single_qubit_matrix_to_phxz(rot_1)) is not None:
                circuit += op_1(self.qubits[0])

            if (op_2 := cirq.single_qubit_matrix_to_phxz(rot_2)) is not None:
                circuit += op_2(self.qubits[1])

            circuit += cirq.measure(sorted(circuit.all_qubits()))

            samples.append(
                Sample(
                    circuit_realization=index,
                    circuit=circuit,
                    data={"num_two_qubit_gates": 2 * depth},
                )
            )
        return samples

    @staticmethod
    def _haar_random_rotation() -> cirq.Gate:
        """Returns:
        Haar randomly sampled SU(2) rotation.
        """
        gate = cirq.single_qubit_matrix_to_phxz(cirq.testing.random_special_unitary(dim=2))
        if gate is None:
            return cirq.I
        return gate

    def _component(self, include_two_qubit_gate: bool) -> cirq.Circuit:
        """Core component of the experimental circuits.

        These circuits that are repeated to create the full circuit. Can optionally include the
        two qubit gate being measured, as is required for the
        first half of the full circuit, but not for the second half.

        The component looks like:
        .. code::

            0: ───R───Q───X───Q───
                      │       │
            1: ───R───Q───X───Q───

        where :code:`R` is a Haar randomly chosen SU(2) rotation
        and :code:`Q-Q` represents the two qubit gate being measured.

        Args:
            include_two_qubit_gate: Whether to include the two qubit gate being measured

        Returns:
            The sub circuit to be repeated when building the full circuit
        """
        # Chose a global rotation
        rotation = self._haar_random_rotation()
        return cirq.Circuit(
            rotation.on(self.qubits[0]),
            rotation.on(self.qubits[1]),
            css.barrier(*self.qubits),
            (
                [
                    self.two_qubit_gate(*self.qubits).with_tags("no_compile"),
                    css.barrier(*self.qubits),
                ]
                if include_two_qubit_gate
                else []
            ),
            cirq.X.on_each(*self.qubits),
            css.barrier(*self.qubits),
            (
                [
                    self.two_qubit_gate(*self.qubits).with_tags("no_compile"),
                    css.barrier(*self.qubits),
                ]
                if include_two_qubit_gate
                else []
            ),
        )

    def _json_dict_(self) -> dict[str, Any]:
        """Converts the experiment to a json-able dictionary that can be used to recreate the
        experiment object. Note that the state of the random number generator is not stored.

        Returns:
            Json-able dictionary of the experiment data.
        """
        return {
            "two_qubit_gate": self.two_qubit_gate,
            **super()._json_dict_(),
        }

    @classmethod
    def _from_json_dict_(
        cls,
        _samples: list[Sample],
        num_circuits: int,
        cycle_depths: Iterable[int],
        two_qubit_gate: cirq.Gate = cirq.CZ,
        **kwargs: Any,
    ) -> Self:
        """Creates a experiment from a dictionary of the data.

        Args:
            dictionary: Dict containing the experiment data.

        Returns:
            The deserialized experiment object.
        """
        kwargs.pop("qubits")  # Number of qubits is fixed for SU2
        return cls(
            num_circuits=num_circuits,
            _samples=_samples,
            cycle_depths=cycle_depths,
            two_qubit_gate=two_qubit_gate,
            **kwargs,
        )



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for cross entropy benchmark experiments."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import cirq
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import seaborn as sns
import tqdm.contrib.itertools

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample

if TYPE_CHECKING:
    import numpy.typing as npt


@dataclass
class XEBResults(QCVVResults):
    """Results from an XEB experiment."""

    _circuit_fidelities: pd.DataFrame | None = None
    _cycle_fidelity_estimate: float | None = None
    """Estimated cycle fidelity."""
    _cycle_fidelity_estimate_std: float | None = None
    """Standard deviation for the cycle fidelity estimate."""

    @property
    def cycle_fidelity_estimate(self) -> float:
        """Estimated cycle fidelity."""
        if self._cycle_fidelity_estimate is None:
            raise self._not_analyzed
        return self._cycle_fidelity_estimate

    @property
    def cycle_fidelity_estimate_std(self) -> float:
        """Standard deviation for the cycle fidelity estimate."""
        if self._cycle_fidelity_estimate_std is None:
            raise self._not_analyzed
        return self._cycle_fidelity_estimate_std

    def _analytical_data(self) -> pd.DataFrame:
        """Create a copy of `self.data` with analytical probabilities in place of experimental data.

        Returns:
            The `pd.DataFrame` containing analytical probabilities.
        """
        assert self.data is not None

        indices = [f"{i:0>{self.num_qubits}b}" for i in range(2**self.num_qubits)]

        analytic_probabilities: list[npt.NDArray[np.float64]] = []
        for _, row in self.data.iterrows():
            sample = self.experiment[row.uuid]
            subcircuit = sample.circuit.copy()
            subcircuit.clear_operations_touching(
                subcircuit.all_qubits().difference(self.qubits), range(len(subcircuit))
            )
            analytic_final_state = cirq.final_state_vector(
                subcircuit,
                qubit_order=self.qubits,
                ignore_terminal_measurements=True,
                dtype=np.complex128,
            )
            analytic_probabilities.append(np.abs(analytic_final_state) ** 2)

        analytical_data = self.data.drop(columns=indices, errors="ignore")
        analytical_data[indices] = analytic_probabilities
        return analytical_data

    def _analyze(self) -> None:
        """Analyse the results and calculate the estimated circuit fidelity.

        Args:
            plot_results (optional): Whether to generate the data plots. Defaults to True.

        Raises:
            RuntimeError: If there is no data stored.

        Returns:
           The final results from the experiment.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot perform analysis.")

        analytical_data = self._analytical_data()

        indices = [f"{i:0>{self.num_qubits}b}" for i in range(2**self.num_qubits)]
        self.data["sum_p(x)p^(x)"] = (self.data[indices] * analytical_data[indices]).sum(axis=1)
        self.data["sum_p(x)p(x)"] = (analytical_data[indices] ** 2).sum(axis=1)

        # Fit a linear model for each cycle depth to estimate the circuit fidelity
        records = []
        for depth in set(self.data.cycle_depth):
            df = self.data[self.data.cycle_depth == depth]
            fit = scipy.stats.linregress(
                x=df["sum_p(x)p(x)"] - 1 / 2**self.num_qubits,
                y=df["sum_p(x)p^(x)"] - 1 / 2**self.num_qubits,
            )

            records.append(
                {
                    "cycle_depth": depth,
                    "circuit_fidelity_estimate": fit.slope,
                    "circuit_fidelity_estimate_std": fit.stderr,
                }
            )

        self._circuit_fidelities = pd.DataFrame(records)

        self._circuit_fidelities["log_fidelity_estimate"] = np.log(
            self._circuit_fidelities.circuit_fidelity_estimate
        )

        # Fit a linear model to the depth ~ log(fidelity) to approximate the cycle fidelity
        cycle_fit = scipy.stats.linregress(
            x=self._circuit_fidelities.cycle_depth,
            y=np.log(self._circuit_fidelities.circuit_fidelity_estimate),
        )
        self._cycle_fidelity_estimate = np.exp(cycle_fit.slope)
        self._cycle_fidelity_estimate_std = self.cycle_fidelity_estimate * cycle_fit.stderr
        self._zero_cycle_fidelity_estimate = np.exp(cycle_fit.intercept)

    def plot_results(
        self,
        filename: str | None = None,
    ) -> plt.Figure:
        """Plot the experiment data and the corresponding fits.

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.

        Returns:
            A single matplotlib figure containing both the linear fit per cycle depth and the decay
            with cycle depth.

        Raises:
            RuntimeError: If there is no data stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot plot results.")

        if self._circuit_fidelities is None:
            raise RuntimeError(
                "No stored dataframe of circuit fidelities. Something has gone wrong."
            )

        fig, axs = plt.subplots(1, 2, figsize=(10, 4.8))
        colours = sns.color_palette("dark:r", n_colors=len(self.data.cycle_depth.unique()))
        for depth, color in zip(sorted(self.data.cycle_depth.unique()), colours):
            sns.regplot(
                data=self.data[self.data.cycle_depth == depth],
                x="sum_p(x)p(x)",
                y="sum_p(x)p^(x)",
                ci=None,
                ax=axs[0],
                color=color,
                label=depth,
            )
        axs[0].legend(title="Cycle depth", bbox_to_anchor=(1.0, 0.5), loc="center left")
        axs[0].set_xlabel(r"$\sum p(x)^2$", fontsize=15)
        axs[0].set_ylabel(r"$\sum p(x) \hat{p}(x)$", fontsize=15)
        axs[0].set_title(r"Linear fit per cycle depth", fontsize=15, wrap=True)

        sns.regplot(
            data=self._circuit_fidelities,
            x="cycle_depth",
            y="circuit_fidelity_estimate",
            ax=axs[1],
            fit_reg=False,
            color="tab:red",
        )
        axs[1].set_xlabel(r"Cycle depth", fontsize=15)
        axs[1].set_ylabel(r"Circuit fidelity", fontsize=15)
        axs[1].set_title(r"Exponential decay of circuit fidelity", fontsize=15, wrap=True)

        # Add fit line
        x = np.linspace(
            self._circuit_fidelities.cycle_depth.min(), self._circuit_fidelities.cycle_depth.max()
        )
        y = self.cycle_fidelity_estimate**x * self._zero_cycle_fidelity_estimate
        y_p = (
            self.cycle_fidelity_estimate + 1.96 * self.cycle_fidelity_estimate_std
        ) ** x * self._zero_cycle_fidelity_estimate
        y_m = (
            self.cycle_fidelity_estimate - 1.96 * self.cycle_fidelity_estimate_std
        ) ** x * self._zero_cycle_fidelity_estimate
        axs[1].plot(x, y, color="tab:red", linewidth=2)
        axs[1].fill_between(x, y_m, y_p, alpha=0.25, color="tab:red", label="95% CI")
        axs[1].legend()

        fig.tight_layout()

        if filename is not None:
            fig.savefig(filename, bbox_inches="tight")

        return fig

    def print_results(self) -> None:
        """Prints the key results data."""
        print(  # noqa: T201
            f"Estimated cycle fidelity: {self.cycle_fidelity_estimate:.5} "
            f"+/- {self.cycle_fidelity_estimate_std:.5}"
        )

    def plot_speckle(self, filename: str | None = None) -> plt.Figure:
        """Creates the speckle plot of the XEB data. See Fig. S18 of
        https://arxiv.org/abs/1910.11333 for an explanation of this plot.

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.

        Returns:
            A matplotlib figure with the speckle plot.

        Raises:
            RuntimeError: If there is no data stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot plot results.")

        indices = [f"{i:0>{self.num_qubits}b}" for i in range(2**self.num_qubits)]

        # Reformat dataframe
        df = pd.melt(
            self.data,
            value_vars=indices,
            id_vars=["cycle_depth", "circuit_realization"],
            var_name="bitstring",
        )

        # Create the axes needed
        fig, axs = plt.subplot_mosaic(
            [[f"P({idx})", "cbar", ".", "Decay"] for idx in indices],
            width_ratios=[1, 0.05, 0.05, 1],
            figsize=(12, 4.8),
        )
        fig.subplots_adjust(hspace=0)

        # Plot the heatmaps
        for k, bitstring in enumerate(indices):
            ax = axs[f"P({bitstring})"]

            data = df[df["bitstring"] == bitstring].pivot(
                index="circuit_realization", columns="cycle_depth", values="value"
            )
            cmap = mpl.colormaps["rocket"]
            # norm = mpl.colors.Normalize(0, 1)  # or vmin, vmax # noqa: ERA001
            sns.heatmap(data, vmin=0, vmax=1, ax=ax, cbar_ax=axs["cbar"], cmap=cmap)
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_yticks([])
            ax.set_yticklabels([])
            plt.text(
                0.99,
                0.90,
                f"P({bitstring})",
                ha="right",
                va="top",
                transform=ax.transAxes,
                color="white",
            )
            if k == 0:
                ax.set_title("Speckle plots")
            else:
                ax.axhline(y=0, linewidth=1.5, color="white", linestyle="--")

            if k == 2**self.num_qubits - 1:
                ax.set_xlabel("Cycle depth")

        # Format colour bar
        axs["cbar"].set_ylabel("Probability")
        axs["cbar"].yaxis.set_label_position("right")
        axs["cbar"].yaxis.tick_left()

        # Plot the decay of purity

        # Calculate the average std of the probability distributions.
        purity_data = (
            df.groupby(by=["cycle_depth"])
            .std(numeric_only=True)
            .reset_index()
            .rename(columns={"value": "sqrt_speckle_purity"})
            .drop(columns=["circuit_realization"])
        )
        # Rescale the purity estimate according to Porter-Thomas distribution
        dim = np.prod([q.dimension for q in self.qubits])
        purity_data["sqrt_speckle_purity"] = purity_data["sqrt_speckle_purity"] * np.sqrt(
            dim**2 * (dim + 1) / (dim - 1)
        )

        # Plot decay
        sns.regplot(
            data=purity_data,
            x="cycle_depth",
            y="sqrt_speckle_purity",
            logx=True,
            ax=axs["Decay"],
        )
        # purity_plot.tight_layout() # noqa: ERA001
        axs["Decay"].set_xlabel(r"Cycle depth")
        axs["Decay"].set_ylabel(r"$\sqrt{\mathrm{Purity}}$")
        axs["Decay"].set_title(r"Purity Decay")
        # Estimate decay coefficient
        purity_fit = scipy.stats.linregress(
            x=purity_data["cycle_depth"],
            y=np.log(purity_data["sqrt_speckle_purity"]),
        )
        # Add label with coefficient estimate
        axs["Decay"].text(
            0.95,
            0.95,
            (
                f"Decay coefficient: {np.exp(purity_fit.slope):4f} "
                f"+/- {np.exp(purity_fit.slope) * purity_fit.stderr:4f}"
            ),
            ha="right",
            va="center",
            transform=axs["Decay"].transAxes,
        )

        if filename is not None:
            fig.savefig(filename, bbox_inches="tight")

        return fig


class XEB(QCVVExperiment[XEBResults]):
    r"""Cross-entropy benchmarking (XEB) experiment.

    The XEB experiment can be used to estimate the combined fidelity of a repeating cycle of gates.
    Cycles are made up of randomly selected single qubit gates and a constant layer of interest.
    This is illustrated as follows:

    For each randomly generated circuit, with a given number of cycle, we compare the
    simulated state probabilities, :math:`p(x)` with those achieved by running the circuit
    on a given target, :math:`\hat{p}(x)`. The fidelity of a circuit containing :math:`d`
    cycles, :math:`f_d` can then be estimated as

    .. math::

        \sum_{x \in \{0, 1\}^n} p(x) \hat{p}(x) - \frac{1}{2^n} =
        f_d \left(\sum_{x \in \{0, 1\}^n} p(x)^2 -  \frac{1}{2^n}\right)

    We can therefore fit a linear model to estimate the value of :math:`f_d`. We the estimate
    the fidelity of the cycle, :math:`f_{\mathrm{cycle}}` as

    .. math::

        f_d = A(f_{cycle})^d

    Thus fitting another linear model to :math:`\log(f_d) \sim d` provides us with an estimate
    of the cycle fidelity.

    For more details see: https://www.nature.com/articles/s41586-019-1666-5
    """

    def __init__(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
        interleaved_layer: cirq.Gate | cirq.OP_TREE | None = cirq.CZ,
        single_qubit_gate_set: Sequence[cirq.Gate] | None = None,
        *,
        random_seed: int | np.random.Generator | None = None,
        _samples: list[Sample] | None = None,
        **kwargs: str,
    ) -> None:
        """Initializes a cross-entropy benchmarking experiment.

        Args:
            num_circuits: Number of circuits to sample.
            cycle_depths: The cycle depths to sample.
            interleaved_layer: The gate or operation(s) to interleave between the single qubit
                gates. If None then no gates are interleaved.
            single_qubit_gate_set: Optional list of single qubit gates to randomly sample from when
                generating random circuits. If not provided defaults to phased X gates with 1/4 pi
                intervals.
            random_seed: An optional seed to use for randomization.
            kwargs: Any additional supported string keyword args.
        """
        if interleaved_layer is None:
            qubits: Sequence[cirq.Qid] = cirq.LineQubit.range(2)

        elif isinstance(interleaved_layer, cirq.Gate):
            qubits = (
                cirq.LineQubit.range(cirq.num_qubits(interleaved_layer))
                if all(d == 2 for d in cirq.qid_shape(interleaved_layer))
                else cirq.LineQid.for_gate(interleaved_layer)
            )
            interleaved_layer = interleaved_layer.on(*qubits)

        elif isinstance(interleaved_layer, (cirq.Operation, cirq.Moment)):
            qubits = sorted(interleaved_layer.qubits)

        else:
            interleaved_layer = cirq.Circuit(interleaved_layer)
            qubits = sorted(interleaved_layer.all_qubits())

        self.interleaved_layer: cirq.OP_TREE | None = interleaved_layer
        """The layer to interleave."""

        self.single_qubit_gate_set: list[cirq.Gate]
        """The single qubit gates to randomly sample from"""

        if single_qubit_gate_set is None:
            self.single_qubit_gate_set = [
                cirq.PhasedXPowGate(exponent=0.5, phase_exponent=phase_exponent)
                for phase_exponent in [0.0, 0.25, 0.5]
            ]
        else:
            self.single_qubit_gate_set = list(single_qubit_gate_set)

        super().__init__(
            qubits=qubits,
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            random_seed=random_seed,
            results_cls=XEBResults,
            _samples=_samples,
            **kwargs,
        )

    def independent_qubit_groups(self) -> list[tuple[cirq.Qid, ...]]:
        """Get all independent subsets of qubits in this experiment.

        Returns:
            A list of disjoint tuples of `cirq.Qid` objects, each of which can be analyzed as an
            independent XEB experiment.
        """
        if self.interleaved_layer:
            qubit_sets = cirq.Circuit(self.interleaved_layer).get_independent_qubit_sets()
            return sorted((tuple(sorted(qubit_set)) for qubit_set in qubit_sets), key=min)

        return [self.qubits]

    ###################
    # Private Methods #
    ###################
    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[Sample]:
        """Build a list of random circuits to perform the XEB experiment with.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different numbers of cycles to include in each circuit.

        Returns:
            The list of experiment samples.
        """
        random_circuits = []
        for k, depth in tqdm.contrib.itertools.product(
            range(num_circuits), cycle_depths, desc="Building circuits"
        ):
            # Choose single-qubit gates, avoiding repeats on the same qubit in sequential layers
            num_choices = len(self.single_qubit_gate_set)
            block_repeats = num_choices > 1
            chosen_gate_indices = np.append(
                self._rng.integers(0, num_choices, size=(1, self.num_qubits)),
                self._rng.integers(block_repeats, num_choices, size=(depth, self.num_qubits)),
                axis=0,
            )
            chosen_gate_indices = np.add.accumulate(chosen_gate_indices, axis=0) % num_choices

            circuit = cirq.Circuit(
                self.single_qubit_gate_set[gate_index].on(qubit)
                for gate_indices_in_layer in chosen_gate_indices
                for gate_index, qubit in zip(gate_indices_in_layer, self.qubits)
            )

            circuit = self._interleave_layer(circuit, self.interleaved_layer)

            random_circuits.append(
                Sample(
                    circuit=circuit + cirq.measure(*self.qubits),
                    data={
                        "circuit_depth": self._count_non_barrier_gates(circuit),
                        "cycle_depth": depth,
                        "interleaved_layer": str(self.interleaved_layer),
                    },
                    circuit_realization=k,
                )
            )

        return random_circuits

    def _json_dict_(self) -> dict[str, Any]:
        """Converts the experiment to a json-able dictionary that can be used to recreate the
        experiment object. Note that the state of the random number generator is not stored.

        Returns:
            Json-able dictionary of the experiment data.
        """
        json_dict = super()._json_dict_()
        json_dict["interleaved_layer"] = self.interleaved_layer
        json_dict["single_qubit_gate_set"] = self.single_qubit_gate_set
        del json_dict["qubits"]

        return json_dict



from __future__ import annotations

# ruff: noqa: D402
import abc
from typing import TYPE_CHECKING, Any

import cirq

import supermarq

if TYPE_CHECKING:
    import qiskit


class Benchmark:
    """Class representing a quantum benchmark application.

    Concrete subclasses must implement the abstract methods ``circuit()`` and
    ``score()``.

    Each instantiation of a `Benchmark` object represents a single, fully defined
    benchmark application. All the relevant parameters for a benchmark should
    be passed in upon creation, and will be used to generate the correct circuit
    and compute the final score.
    """

    @abc.abstractmethod
    def circuit(self) -> cirq.Circuit | list[cirq.Circuit]:
        """Returns the quantum circuit(s) corresponding to the current benchmark parameters."""

    def cirq_circuit(self) -> cirq.Circuit | list[cirq.Circuit]:
        """Returns:
        The cirq circuit(s) corresponding to the current benchmark parameters.
        """
        return self.circuit()

    def qiskit_circuit(self) -> qiskit.QuantumCircuit | list[qiskit.QuantumCircuit]:
        """Returns:
        The qiskit circuit(s) corresponding to the current benchmark parameters.
        """
        cirq_circuit = self.cirq_circuit()
        if isinstance(cirq_circuit, cirq.Circuit):
            return supermarq.converters.cirq_to_qiskit(cirq_circuit)
        return [supermarq.converters.cirq_to_qiskit(c) for c in cirq_circuit]

    @abc.abstractmethod
    def score(self, counts: Any) -> float:
        """Returns a normalized [0,1] score reflecting device performance.

        Args:
            counts: Dictionary(s) containing the measurement counts from execution.
        """
